using System.Diagnostics;
using Godot;

public class Game : Spatial {
    private enum ActionMode {
        Default,
        Building
    }

    private ActionMode _actionMode = ActionMode.Default;

    private ImageTexture _blackWhitePatternTexture;
    private Camera _camera;
    private Vector3 _cameraOffset;
    private Label _centerLabel;

    private HexCell _currentTile;

    // ui elements
    private Label _framesPerSecondLabel;
    private Control _gameUi;
    private Label _woodCountLabel;
    private Label _goldCountLabel;
    private TextureRect _heightTextureRect;
    private Button _walkActionButton;
    private Button _buildActionButton;

    // other members
    private HexGrid _hexGrid;
    private InteractionSystem _interactionSystem;
    private Label _mouseTileCubeLabel;
    private Label _mouseTileAxialLabel;
    private Spatial _mouseTileHighlight;
    private Label _mouseTileOffsetLabel;
    private Label _mouseWorldLabel;
    private Player _player;

    // scene nodes
    private Spatial _tileHighlight;

    // Resources
    private ShaderMaterial _tileMaterial;
    private Label _tileOffsetLabel;
    private World _world;
    private TextureRect _worldTextureRect;
    private readonly PackedScene _workbenchScene = GD.Load<PackedScene>("res://entities/Workbench.tscn");

    // Called when the node enters the scene tree for the first time.
    public override void _Ready() {
        // debugStatsContainer
        Container debugStatsContainer = (Container)FindNode("DebugStatsContainer");

        _framesPerSecondLabel = debugStatsContainer.GetNode<Label>("fps_label");
        _centerLabel = debugStatsContainer.GetNode<Label>("center_label");
        _tileOffsetLabel = debugStatsContainer.GetNode<Label>("tile_offset_label");
        _mouseWorldLabel = debugStatsContainer.GetNode<Label>("mouse_world_label");
        _mouseTileOffsetLabel = debugStatsContainer.GetNode<Label>("mouse_tile_offset_label");
        _mouseTileCubeLabel = debugStatsContainer.GetNode<Label>("mouse_tile_cube_label");
        _mouseTileAxialLabel = debugStatsContainer.GetNode<Label>("mouse_tile_axial_label");

        // UI elements
        Container worldGeneratorWidget = (Container)FindNode("WorldGeneratorWidget");
        _worldTextureRect = worldGeneratorWidget.GetNode<TextureRect>("WorldTextureRect");
        _heightTextureRect = worldGeneratorWidget.GetNode<TextureRect>("HeightTextureRect");
        _gameUi = (Control)FindNode("GameUI");
        _woodCountLabel = _gameUi.GetNode<Label>("TopRight/WoodCount");
        Debug.Assert(_woodCountLabel != null);
        _goldCountLabel = _gameUi.GetNode<Label>("TopRight/GoldCount");
        Debug.Assert(_goldCountLabel != null);
        _walkActionButton = _gameUi.GetNode<Button>("Actions/WalkActionButton");
        Debug.Assert(_walkActionButton != null);
        _buildActionButton = _gameUi.GetNode<Button>("Actions/BuildActionButton");
        Debug.Assert(_buildActionButton != null);

        // scene nodes
        _tileHighlight = GetNode<Spatial>("TileHighlight");
        _mouseTileHighlight = GetNode<Spatial>("MouseTileHighlight");

        _player = GetNode<Player>("Player");
        _camera = (Camera)FindNode("Camera");
        _cameraOffset = _camera.GlobalTranslation - _player.GlobalTranslation;

        _world = (World)FindNode("World");

        // resources        
        _tileMaterial = GD.Load<ShaderMaterial>("materials/HexTileTextureLookup.tres");
        Debug.Assert(_tileMaterial != null);

        _blackWhitePatternTexture = new ImageTexture();
        Image image = new();
        image.Load("assets/4x4checker.png");
        _blackWhitePatternTexture.CreateFromImage(image, (uint)(Texture.FlagsEnum.Mipmaps | Texture.FlagsEnum.Repeat));

        // other members
        _currentTile = new HexCell();
        _hexGrid = new HexGrid();
        _interactionSystem = GetNode<InteractionSystem>("InteractionSystem");
        Debug.Assert(_interactionSystem != null);

        // connect signals
        _player.TaskQueueComponent.Connect("StartInteraction", _interactionSystem,
            nameof(_interactionSystem.OnStartInteraction));
        _player.Connect("WoodCountChanged", this, nameof(OnWoodCountChanged));
        _player.Connect("GoldCountChanged", this, nameof(OnGoldCountChanged));
        _world.Connect("TileClicked", this, nameof(OnTileClicked));
        _world.Connect("TileHovered", this, nameof(OnTileHovered));
        _world.Connect("OnWorldViewTileTypeImageChanged", this, nameof(OnWorldViewTileTypeImageChanged));
        _world.Connect("OnHeightmapImageChanged", this, nameof(OnHeightmapImageChanged));
        _walkActionButton.Connect("pressed", this, nameof(OnWalkActionPressed));
        _buildActionButton.Connect("pressed", this, nameof(OnBuildActionPressed));

        // register entity events
        foreach (Node node in GetNode("Entities").GetChildren()) {
            if (node.HasSignal("EntityClicked")) {
                node.Connect("EntityClicked", this, nameof(OnEntityClicked));
            }
        }

        _world.Connect("EntityClicked", this, nameof(OnEntityClicked));

        // perform dependency injection
        //_streamContainer.SetWorld(_tileWorld);Clicked
        WorldInfoComponent worldInfoComponent = _player.GetNode<WorldInfoComponent>("WorldInfo");

        UpdateCurrentTile();

        StartNewGame(2, 12);
    }


    public override void _Input(InputEvent inputEvent) {
        if (inputEvent.IsAction("Forward")) {
            GD.Print("Forward");
        }

        if (inputEvent.IsAction("Back")) {
            GD.Print("Back");
        }
    }

    public void StartNewGame(int seed, int chunkSize) {
        _world.Seed = seed;
        _world.ChunkSize = chunkSize;
        _world.InitNoiseGenerator();

        ResetGame();

        _world.UpdateCenterChunkFromPlaneCoord(Vector2.Zero);
    }

    public void ResetGame() {
        _player.GlobalTranslation = Vector3.Zero;
        _player.PlaneAngle = -Mathf.Pi * 0.5f;

        _world.Reset();
    }

    public void UpdateCurrentTile() {
        // cast a ray from the camera to center
        Vector3 cameraNormal = _camera.ProjectRayNormal(_camera.GetViewport().Size * 0.5f);
        Vector3 cameraPosition = _camera.ProjectRayOrigin(_camera.GetViewport().Size * 0.5f);
        Vector3 cameraDir = cameraNormal - cameraPosition;

        Vector3 centerCoord;

        if (Mathf.Abs(cameraDir.y) > Globals.EpsPosition) {
            centerCoord = cameraPosition + cameraNormal * (-cameraPosition.y / cameraNormal.y);
        } else {
            centerCoord = _camera.GlobalTranslation;
            centerCoord.y = 0;
        }

        _world.SetCenterPlaneCoord(new Vector2(_player.GlobalTranslation.x, _player.GlobalTranslation.z));

        _currentTile = _hexGrid.GetHexAt(new Vector2(centerCoord.x, centerCoord.z));

        _centerLabel.Text = centerCoord.ToString("F3");
        _tileOffsetLabel.Text = _currentTile.OffsetCoords.ToString();
    }


    public override void _Process(float delta) {
        _framesPerSecondLabel.Text = Engine.GetFramesPerSecond().ToString();

        UpdateCurrentTile();

        Transform tileHighlightTransform = Transform.Identity;
        Vector2 currentTileCenter = _hexGrid.GetHexCenter(_currentTile);
        tileHighlightTransform.origin.x = currentTileCenter.x;
        tileHighlightTransform.origin.z = currentTileCenter.y;
        _tileHighlight.Transform = tileHighlightTransform;

        Transform cameraTransform = _camera.Transform;
        cameraTransform.origin = _player.GlobalTranslation + _cameraOffset;
        _camera.Transform = cameraTransform;
    }

    public void OnAreaInputEvent(Node camera, InputEvent inputEvent, Vector3 position, Vector3 normal,
        int shapeIndex) {
        HexCell cellAtCursor = _hexGrid.GetHexAt(new Vector2(position.x, position.z));
        Transform highlightTransform = Transform.Identity;

        _mouseWorldLabel.Text = position.ToString("F3");
        _mouseTileOffsetLabel.Text = cellAtCursor.OffsetCoords.ToString("N");
        _mouseTileCubeLabel.Text = cellAtCursor.CubeCoords.ToString("N");
        _mouseTileAxialLabel.Text = cellAtCursor.AxialCoords.ToString("N");
        _mouseTileHighlight.Transform = highlightTransform;
    }


    public void OnTileClicked(HexTile3D tile) {
        if (_player == null) {
            return;
        }

        if (_actionMode == ActionMode.Building) {
            Workbench workbench = (Workbench)_workbenchScene.Instance();

            workbench.Connect("EntityClicked", this, nameof(OnEntityClicked));
            workbench.Transform = tile.GlobalTransform;

            AddChild(workbench);
            _world.MarkCellUnwalkable(tile.Cell);

            return;
        }

        if (_player.InteractionComponent != null) {
            _player.InteractionComponent.EmitSignal("InteractionEnd");
        }

        _player.TaskQueueComponent.Reset();
        _player.TaskQueueComponent.Queue.Enqueue(new TaskQueueComponent.NavigationTask(
            new NavigationPoint(tile.GlobalTranslation)));
    }

    public void OnTileHovered(HexTile3D tile) {
        Transform highlightTransform = tile.GlobalTransform;
        _mouseTileHighlight.Transform = highlightTransform;
        _mouseWorldLabel.Text = highlightTransform.origin.ToString("F3");
        _mouseTileOffsetLabel.Text = tile.OffsetCoords.ToString("N");
        _mouseTileCubeLabel.Text = tile.Cell.CubeCoords.ToString("N");

        _player.NavigationComponent.FindPath(_player, _player.GlobalTranslation, tile.GlobalTranslation);
    }

    public void OnEntityClicked(Entity entity) {
        GD.Print("Clicked on entity at " + entity.GlobalTranslation);

        Spatial mountPoint = (Spatial)entity.FindNode("MountPoint");
        if (mountPoint != null) {
            _player.TaskQueueComponent.Reset();
            _player.TaskQueueComponent.Queue.Enqueue(new TaskQueueComponent.NavigationTask(
                new NavigationPoint(mountPoint.GlobalTransform)));
            _player.TaskQueueComponent.Queue.Enqueue(new TaskQueueComponent.InteractionTask(entity));
        }
    }


    public void ResetGameState() {
        Transform playerStartTransform = Transform.Identity;
        playerStartTransform.origin.y = 0;
        _player.Transform = playerStartTransform;
        _player.TaskQueueComponent.Reset();
        _player.NavigationComponent.PlanDirectPath(_player, playerStartTransform.origin, playerStartTransform.origin,
            playerStartTransform.basis.Quat());

        _woodCountLabel.Text = "0";
        _goldCountLabel.Text = "0";
        _actionMode = ActionMode.Default;

        foreach (Spatial entity in GetNode("Entities").GetChildren()) {
            Transform entityTransform = entity.Transform;
            Vector2 entityPlanePos = new(entityTransform.origin.x, entityTransform.origin.z);
            Vector2 entityOffsetCoordinates = _hexGrid.GetHexAt(entityPlanePos).OffsetCoords;
            entityTransform.origin.y = 0;
            entity.Transform = entityTransform;
        }
    }

    private void OnHeightmapImageChanged(Image heightmapImage) {
        ImageTexture newHeightmapTexture = new();
        newHeightmapTexture.CreateFromImage(heightmapImage,
            (uint)(Texture.FlagsEnum.Mipmaps | Texture.FlagsEnum.Repeat));

        _heightTextureRect.Texture = newHeightmapTexture;
    }

    private void OnWorldViewTileTypeImageChanged(Image viewTileTypeImage) {
        ImageTexture newWorldTexture = new();
        newWorldTexture.CreateFromImage(viewTileTypeImage,
            (uint)(Texture.FlagsEnum.Mipmaps | Texture.FlagsEnum.Repeat));

        _worldTextureRect.Texture = newWorldTexture;

        _tileMaterial.SetShaderParam("MapAlbedoTexture", newWorldTexture);
        _tileMaterial.SetShaderParam("TextureSize", (int)newWorldTexture.GetSize().x);
        _tileMaterial.SetShaderParam("CoordinateOffsetU", (int)_world.WorldTextureCoordinateOffset.x);
        _tileMaterial.SetShaderParam("CoordinateOffsetV", (int)_world.WorldTextureCoordinateOffset.y);
    }

    public void OnWoodCountChanged(int woodCount) {
        AnimationPlayer animationPlayer = _woodCountLabel.GetNode<AnimationPlayer>("AnimationPlayer");
        _woodCountLabel.Text = woodCount.ToString();
        animationPlayer.CurrentAnimation = "FlashLabel";
        animationPlayer.Seek(0);
        animationPlayer.Play();
    }

    public void OnGoldCountChanged(int goldCount) {
        AnimationPlayer animationPlayer = _goldCountLabel.GetNode<AnimationPlayer>("AnimationPlayer");
        _goldCountLabel.Text = goldCount.ToString();
        animationPlayer.CurrentAnimation = "FlashLabel";
        animationPlayer.Seek(0);
        animationPlayer.Play();
    }

    public void OnWalkActionPressed() {
        _actionMode = ActionMode.Default;
    }

    public void OnBuildActionPressed() {
        _actionMode = ActionMode.Building;
    }
}